import time
from typing import AsyncGenerator
from typing import Iterable
from typing import Literal
from typing import TYPE_CHECKING

import httpx

from aegisx.ext.oauth.models import ClientConfiguration
from aegisx.ext.oauth.models import Grant
from aegisx.ext.oauth.models import TokenResponse
from aegisx.ext.oauth.protocols import IClientRepository
from aegisx.ext.oauth.types import AccessTokenType
if TYPE_CHECKING:
    from aegisx.ext.oauth.client import Client


class BaseResourceServerAuth(httpx.Auth):
    """Base class for all OAuth 2.x/OpenID Connect authentication flows
    with resource servers.
    """
    config: ClientConfiguration
    ephemeral_port: int
    grant: Grant | None
    leeway: int = 0
    refresh_status_codes: set[int]
    response_mode: str
    response_type: str
    scope: set[str]


    def __init__(
        self,
        name: str,
        config: ClientConfiguration,
        *,
        repo: IClientRepository,
        scope: Iterable[str] | None = None,
        refresh_status_codes: set[int] = {401, 403},
        response_type: Literal['code', 'id_token', 'code id_token', 'code id_token token'] = 'code',
        response_mode: Literal['query', 'query.jwt'] = 'query',
        ephemeral_port: int = 0
    ):
        self.config = config
        self.ephemeral_port = ephemeral_port
        self.grant = None
        self.name = name
        self.refresh_status_codes = refresh_status_codes
        self.response_mode = response_mode
        self.response_type = response_type
        self.repo = repo
        self.scope = set(scope or [])

    def authenticate_request(self, request: httpx.Request) -> None:
        """Authenticate a request using the access token."""
        assert self.grant
        match self.grant.token_type:
            case AccessTokenType.BEARER:
                assert self.grant.access_token
                request.headers['Authorization'] = f'Bearer {self.grant.access_token}'
            case _:
                raise NotImplementedError(
                    f"Tokens of type {self.grant.token_type} are not implemented."
                )

    def client_factory(self) -> 'Client':
        from aegisx.ext.oauth.client import Client # TODO
        return Client.fromconfig(self.config)

    def get_ephemeral_port(self) -> int:
        if not self.ephemeral_port:
            raise NotImplementedError
        return self.ephemeral_port

    def is_invalid(self, response: httpx.Response):
        """Return a boolean indicating if the access token is
        expired, invalid or otherwise not usable.
        """
        # Technically not conforming to spec, but not every resource
        # server conforms to the spec.
        return response.status_code in self.refresh_status_codes

    def must_refresh(self, request: httpx.Request, now: int | None = None) -> bool:
        """Return a boolean indicating if the access token must be
        refreshed.
        """
        if not self.grant or not self.grant.expires_in:
            # If there is no grant or the server did not send the "expires_in"
            # parameter, we do not know if we must refresh.
            return False
        now = int(now or time.time())
        return (now - self.grant.obtained - self.leeway) > self.grant.expires_in

    async def async_auth_flow(
        self,
        request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        if self.grant is None:
            self.grant = await self.repo.grant(self.name)
        if not self.grant or not self.grant.access_token or self.must_refresh(request):
            await self.obtain(request)
        self.authenticate_request(request)

        assert self.grant is not None
        assert self.grant.access_token is not None
        response = yield request
        if self.is_invalid(response):
            await self.obtain(request)
            self.authenticate_request(request)
            yield request

    async def authorize(self) -> None:
        raise NotImplementedError(
            f'{type(self).__name__} does not support the authorization code flow.'
        )

    async def obtain(self, request: httpx.Request) -> None:
        """Obtain a new access token."""
        if not self.grant or not self.grant.refresh_token:
            await self.authorize()
            return
        await self.refresh(request)

    async def refresh(self, request: httpx.Request) -> None:
        """Refresh the current access token."""
        if not self.grant:
            raise TypeError(f'Grant {self.name} does not exist.')
        if self.grant and not self.grant.refresh_token:
            raise TypeError(
                f'Grant "{self.grant.grant_type}" did not provide a '
                'refresh token.'
            )
        async with self.client_factory() as client:
            response = await client.refresh(self.grant.refresh_token)
            if response.is_error():
                # Fetch the grant from the repository as another
                # caller might have expired this access token,
                # since BaseResourceServerAuth instances can be
                # long-lived (application scoped).
                grant = await self.repo.grant(self.name)
                if grant and grant == self.grant:
                    self.grant = grant
                else:
                    raise NotImplementedError
            else:
                self.grant = await self.process_response(response)

    async def process_response(self, response: TokenResponse) -> Grant:
        if response.is_error():
            raise TypeError('Error responses can not be processed.')
        grant = Grant(
            name=self.name,
            grant_type='refresh_token',
            issuer=self.config.metadata.issuer,
            obtained=int(time.time()),
            response=response,
            scope=self.scope
        )
        await self.repo.persist(grant, name=self.name, config=self.config)
        return grant

    def __repr__(self):
        return f"{type(self).__name__}(name='{self.name}')"


import hashlib
from typing import cast
from typing import ClassVar
from typing import Literal
from typing import TypeVar

import pydantic
from aegisx.types import SpaceSeparatedSet
from aegisx.ext.jose.models import JSONWebToken
from libcanonical.types import Base64
from libcanonical.types import EmailAddress
from libcanonical.types import HTTPResourceLocator

from aegisx.ext.oauth.types import AuthenticationMethodReferenceLiteral
from aegisx.ext.oauth.types import OIDCValidationContext


R = TypeVar('R')


class OIDCToken(JSONWebToken):
    supported_algorithms: ClassVar[set[str]] = {
        'HS256', 'HS384', 'HS512',
        'RS256', 'RS384', 'RS512',
        'PS256', 'PS384', 'PS512',
        'ES256', 'ES384', 'ES512',
    }

    acr: HTTPResourceLocator | Literal['0'] = pydantic.Field(
        default='0',
        title="Authentication Context Class Reference (ACR)",
        description=(
            "String specifying an Authentication Context Class Reference (ACR) value "
            "that identifies the Authentication Context Class (ACC) that the "
            "authentication performed satisfied. The value `0` indicates the end-"
            "user authentication did not meet the requirements of ISO/IEC 29115 level "
            "1. For historic reasons, the value `0` is used to indicate that there is "
            "no confidence that the same person is actually there. Authentications with "
            "level 0 SHOULD NOT be used to authorize access to any resource of any "
            "monetary value. An absolute URI or an RFC 6711 registered name SHOULD "
            "be used as the `acr` value; registered names MUST NOT be used with a "
            "different meaning than that which is registered. Parties using this "
            "claim will need to agree upon the meanings of the values used, which "
            "may be context specific. The `acr` value is a case-sensitive string."
        )
    )

    amr: list[AuthenticationMethodReferenceLiteral] = pydantic.Field(
        default_factory=list,
        title="Authentication methods",
        description=(
            "JSON array of strings that are identifiers for authentication "
            "methods used in the authentication."
        )
    )

    at_hash: Base64 | None = pydantic.Field(
        default=None,
        title="Access token hash",
        description=(
            "The base64url encoding of the left-most half of the hash of the octets of the "
            "ASCII representation of the `access_token` value, where the hash algorithm used "
            "is the hash algorithm used in the `alg` header parameter of the ID Token's JOSE "
            "header. For instance, if the alg is `RS256`, hash the `access_token` value with "
            "SHA-256, then take the left-most 128 bits and base64url-encode them. The "
            "`at_hash` value is a case-sensitive string."
        )
    )

    azp: HTTPResourceLocator | str | None = pydantic.Field(
        default=None,
        title="Authorized party",
        description=(
            "The party to which the ID Token was issued. If present, it "
            "MUST contain the OAuth 2.0 Client ID of this party. The `azp` "
            "value is a case-sensitive string containing a string or URI value."
        )
    )

    auth_time: int | None = pydantic.Field(
        default=None,
        title="Authenticated",
        description=(
            "Time when the end-user authentication occurred. Its value is a JSON "
            "number representing the number of seconds from 1970-01-01T00:00:00Z "
            "as measured in UTC until the date/time. When a `max_age` request is "
            "made or when `auth_tim`e` is requested as an essential claim, then t"
            "his claim is present; otherwise, its inclusion is optional."
        )
    )

    c_hash: Base64 | None = pydantic.Field(
        default=None,
        title="Code hash",
        description=(
            "The base64url encoding of the left-most half of the hash of the octets "
            "of the ASCII representation of the `code` value, where the hash algorithm "
            "used is the hash algorithm used in the `alg` Header Parameter of the ID "
            "Token's JOSE Header. For instance, if the alg is HS512, hash the `code` "
            "value with SHA-512, then take the left-most 256 bits and base64url-encode "
            "them. The `c_hash` value is a case-sensitive string. If the ID Token is "
            "issued from the authorization endpoint with a code, which is the case for "
            "the `response_type` values `code id_token` and `code id_token token`, "
            "this parameter is REQUIRED; otherwise, its inclusion is OPTIONAL."
        )
    )

    email: EmailAddress | None = pydantic.Field(
        default=None,
        title="Email",
        description=(
            "The end-user's preferred e-mail address. Its value MUST conform to "
            "the RFC 5322 addr-spec syntax."
        )
    )

    email_verified: bool = pydantic.Field(
        default=False,
        title="Email verified?",
        description=(
            "Is `true` if the end-user's e-mail address has been verified; otherwise `false`. "
            "When `email_verified` is `true`, this means that the authorization server took "
            "affirmative steps to ensure that this e-mail address was controlled by the end-"
            "user at the time the verification was performed."
        )
    )

    family_name: str | None = pydantic.Field(
        default=None,
        title="Family name",
        description=(
            "The surname(s) or last name(s) of the end-user. Note that in some "
            "cultures, people can have multiple family names or no family name; "
            "all can be present, with the names being separated by space characters."
        )
    )

    given_name: str | None = pydantic.Field(
        default=None,
        title="Given name",
        description=(
            "The given name(s) or first name(s) of the end-user. Note that in "
            "some cultures, people can have multiple given names; all can be "
            "present, with the names being separated by space characters."
        )
    )

    name: str | None = pydantic.Field(
        default=None,
        title="Name",
        description=(
            "The end-user's full name in displayable form including all name "
            "parts, possibly including titles and suffixes, ordered according "
            "to the end-user's locale and preferences."
        )
    )

    nonce: str | None = pydantic.Field(
        default=None,
        title="Nonce",
        description=(
            "String value used to associate a client session with an ID Token, and "
            "to mitigate replay attacks. The value is passed through unmodified "
            "from the authorization request to the ID Token. If present in the "
            "ID Token, Clients must verify that the `nonce` claim value is equal "
            "to the value of the `nonce` parameter sent in the authorization request."
        )
    )

    picture: HTTPResourceLocator | None = pydantic.Field(
        default=None,
        title="Picture",
        description=(
            "The URL of the end-user's profile picture. This URL MUST refer "
            "to an image file (for example, a PNG, JPEG, or GIF image file), "
            "rather than to a web page containing an image. Note that this "
            "URL SHOULD specifically reference a profile photo of the end-user "
            "suitable for displaying when describing the end-user, rather than "
            "an arbitrary photo taken by the End-User."
        )
    )

    preferred_username: str | None = pydantic.Field(
        default=None,
        title="Preferred username",
        description=(
            "Shorthand name by which the end-user wishes to be referred to at the "
            "client, such as `janedoe` or `j.doe`. This value MAY be any valid JSON "
            "string including special characters such as @, /, or whitespace. The "
            "client MUST NOT rely upon this value being unique"
        )
    )

    sid: str | None = pydantic.Field(
        default=None,
        title="Session ID",
        description=(
            "The end-users' session identifier. See OpenID Connect Front-Channel "
            "Logout 1.0, Section 3."
        )
    )

    # Non Standard
    typ: str | None = pydantic.Field(
        default=None,
        title="Type"
    )

    hd: str | None = pydantic.Field(
        default=None,
        title="HD",
        description=(
            "Streamline the login process for accounts owned by a Google Cloud "
            "organization. By including the Google Cloud organization domain "
            "(for example, `mycollege.edu`), you can indicate that the account "
            "selection UI should be optimized for accounts at that domain. To "
            "optimize for Google Cloud organization accounts generally instead "
            "of just one Google Cloud organization domain, set a value of an "
            "asterisk (`*`): `hd=*.`."
        )
    )

    @property
    def scope(self) -> SpaceSeparatedSet:
        return SpaceSeparatedSet()

    @pydantic.model_validator(mode='after')
    def validate_oidc(
        self,
        info: pydantic.ValidationInfo,
    ):
        # It is assumed here that all tokens are issued through the
        # authorization code or hybrid flow.
        ctx = cast(OIDCValidationContext | None, info.context)
        if ctx is not None and ctx.get('client_id'): # TODO
            assert set(ctx.keys()) >= {'client_id', 'grant', 'jws', 'metadata', 'now', 'params'}
            now = ctx['now']
            if not ctx['grant'].access_token:
                raise ValueError(
                    'an OpenID Connect ID Token must be validated '
                    'against an access token.'
                )
            if len(ctx['jws'].signatures) > 1:
                raise ValueError('OpenID ID Tokens can not have multiple signers.')
            if len(ctx['jws'].signatures) == 0:
                raise ValueError('OpenID ID Tokens must be signed.')

            # For now only algorithms that have a hash are supported.
            # Other algorithms, such as EdDSA are pending a decision,
            # see https://bitbucket.org/openid/connect/issues/1125/_hash-algorithm-for-eddsa-id-tokens
            alg = list(ctx['jws'].algorithms)[0]
            if alg not in self.supported_algorithms:
                raise ValueError(f'unsupported signature algorithm: {alg}')
            assert alg.dig is not None

            # Validate at_hash and c_hash
            l = (int(alg[-3:]) // 8) // 2
            h = hashlib.new(alg.dig)
            h.update(str.encode(ctx['grant'].access_token, 'ascii'))
            if not self.at_hash:
                raise ValueError('the "at_hash" claim MUST be present.')
            if l != len(self.at_hash):
                raise ValueError(f'invalid "at_hash" length: {len(self.at_hash)}.')
            if bytes(self.at_hash) != h.digest()[:l]:
                raise ValueError('digest mismatch between "at_hash" and "access_token".')

            if self.c_hash:
                h = hashlib.new(alg.dig)
                h.update(str.encode(ctx['grant'].access_token, 'ascii'))
                if l != len(self.c_hash):
                    raise ValueError(f'invalid "c_hash" length: {len(self.c_hash)}.')
                if bytes(self.c_hash) != h.digest()[:l]:
                    raise ValueError('digest mismatch between "c_hash" and "access_token".')

            if self.azp and self.azp != ctx['client_id']:
                raise ValueError('requesting client is not the authorized party.')

            if ctx['client_id'] not in self.aud:
                raise ValueError('requesting client is not the audience.')

            if ctx['params'].max_age:
                # Presence of this attribute is guaranteed by the "required"
                # specification in the validation context.
                assert self.iat
                if (now - self.iat) > ctx['params'].max_age:
                    raise ValueError('"iat" is too far in the past.')

            if ctx['params'].nonce != self.nonce:
                raise ValueError('the "nonce" claim does not match the request parameter.')

        return self


import contextlib
import http
import http.server
import os
import urllib.parse
import webbrowser
from threading import Event
from threading import Thread
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

from aegisx.ext.oauth.types import AccessTokenType
from ._baseresourceserver import BaseResourceServerAuth
if TYPE_CHECKING:
    from aegisx.ext.oauth.client import Client


C = TypeVar('C', bound='Client')

SUCCESS_RESPONSE = b'You can now close this window.'


class InteractiveAuth(BaseResourceServerAuth):
    """Interactive authentication where the resource owner is redirected
    to the authorization endpoint.
    """
    access_token: str | None = None
    expires_in : int | None = None
    obtained: int | None
    leeway: int = 15
    refresh_token: str | None = None
    refresh_status_codes: set[int] = {401, 403}
    response_type: str
    result: urllib.parse.ParseResult | None = None
    token_type: AccessTokenType | None = None

    class request_handler(http.server.SimpleHTTPRequestHandler):
        auth: 'InteractiveAuth'
        event: Event
        ephemeral_port: int

        def do_GET(self) -> None:
            content = "You can now close this window."
            self.send_response(200)
            self.send_header('Content-Type', 'text/plain')
            try:
                p = urllib.parse.urlparse(f'http://127.0.0.1:{self.ephemeral_port}{self.path}')
                response = self.auth.on_redirected(p)
                self.send_header('Content-Length', str(len(response)))
                self.end_headers()
                self.wfile.write(response)
            except Exception:
                response = b'Internal server error'
                self.send_header('Content-Length', str(len(response)))
                self.wfile.write(str.encode(content))

        def log_message(self, format: str, *args: Any) -> None:
            pass

    def on_redirected(self, result: urllib.parse.ParseResult | None) -> bytes:
        self.result = result
        self.event.set()
        return SUCCESS_RESPONSE

    def wait(self):
        self.event.wait()
        assert self.result
        return self.result

    @contextlib.contextmanager
    def redirect_endpoint(self, port: int):
        self.event = Event()
        server = self.server_factory(port)
        thread = Thread(target=server.serve_forever, daemon=True)
        thread.start()
        yield
        server.shutdown()
        self.result = None

    def server_factory(self, port: int):
        return http.server.ThreadingHTTPServer(
            server_address=('127.0.0.1', port),
            RequestHandlerClass=type(
                'RequestHandler',
                (self.request_handler,),
                {
                    'auth': self,
                    'event': self.event,
                    'ephemeral_port': port
                }
            )
        )

    async def authorize(self) -> None:
        port = self.get_ephemeral_port()
        redirect_uri = f'http://127.0.0.1:{port}'
        state = bytes.hex(os.urandom(16))
        async with self.client_factory() as client:
            request, url = client.authorize_url(
                self.response_type,
                redirect_uri=redirect_uri,
                state=state,
                scope=self.scope,
                response_mode=self.response_mode
            )
            with self.redirect_endpoint(port):
                webbrowser.open(url)
                result = self.wait()
            response = await client.on_redirected(result)
            if response.is_error():
                response.fatal()
            token = await client.obtain(request, response)
            if token.is_error():
                token.fatal()
            self.grant = await self.process_response(token)


import time
from typing import Any
from typing import TYPE_CHECKING

from aegisx.ext.jose import JSONWebKeySet
from aegisx.ext.jose import TokenValidator

from aegisx.ext.oauth.const import OICD_STANDARD_CLAIMS
from aegisx.ext.oauth.models import OIDCToken
from aegisx.ext.oauth.models import ServerMetadata
from aegisx.ext.oauth.models import AuthorizationRequestParameters
from aegisx.ext.oauth.models import TokenResponse
if TYPE_CHECKING:
    from ._client import Client


class OIDCTokenValidator(TokenValidator[OIDCToken]):

    def __init__(
        self,
        client: 'Client',
        metadata: ServerMetadata,
        params: AuthorizationRequestParameters,
        grant: TokenResponse,
        jwks: JSONWebKeySet | None = None
    ):
        assert metadata.is_discovered()
        jwks = jwks or metadata.jwks
        self.client = client
        self.grant = grant
        self.metadata = metadata
        self.params = params

        required = set(OICD_STANDARD_CLAIMS)
        if params.max_age:
            required.add('auth_time')
        if params.nonce:
            required.add('nonce')
        if params.claims and params.claims.id_token:
            required.update(params.claims.id_token.requested())
        super().__init__(
            OIDCToken,
            issuer=metadata.issuer,
            audience={client.client_id},
            jwks=jwks,
            required=required
        )

    def get_context(self) -> dict[str, Any]:
        ctx = super().get_context()
        ctx.update({
            'client_id': self.client.client_id,
            'grant': self.grant,
            'metadata': self.metadata,
            'now': int(time.time()),
            'params': self.params
        })
        return ctx


import collections
import functools
import inspect
import re
from inspect import Parameter
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Iterable
from typing import TypeVar
from typing import Union

import fastapi
import pydantic
from aegisx.ext.iam import AnonymousSubject
from aegisx.ext.iam import AuthenticatedSubject
from aegisx.ext.iam import AuthorizationContext
from aegisx.ext.iam import IAMPolicy
from aegisx.ext.jose import JSONWebToken
from aegisx.ext.jose import JWKSResolver

from ._accesstokenbearer import AccessTokenBearer


T = TypeVar('T', default=JSONWebToken, bound=JSONWebToken)


class RequestContext(Generic[T]):
    __bearer_class__: ClassVar[type[AccessTokenBearer[T]]] = AccessTokenBearer[T] # type: ignore
    subject_claims: ClassVar[set[str]] = {'sub', 'email', 'email_verified'}
    subjects: pydantic.TypeAdapter[AnonymousSubject | AuthenticatedSubject]

    @functools.cached_property
    def __signature__(self) -> inspect.Signature:
        sig = inspect.signature(self.authorize)
        params = collections.OrderedDict(sig.parameters.items())
        params['subject'] = Parameter(
            kind=params['subject'].kind,
            name=params['subject'].name,
            annotation=params['subject'].annotation,
            default=fastapi.Depends(
                self.__bearer_class__(
                    types=Union[tuple(self.token_types)],
                    issuers=self.issuers,
                    jwks_resolver=self.jwks_resolver,
                    subject_factory=self.authenticate,
                    max_age=self.max_token_age
                )
            )
        )
        return sig.replace(parameters=list(params.values()))

    def __init__(
        self,
        token_types: Iterable[type[JSONWebToken]],
        *,
        issuers: set[str] | re.Pattern[str],
        domains: set[str],
        policy: IAMPolicy | None = None,
        max_token_age: int = 0
    ):
        self.domains = domains
        self.issuers = issuers
        self.jwks_resolver = JWKSResolver(domains=self.domains)
        self.max_token_age = max_token_age
        self.token_types = token_types
        self.policy = policy or IAMPolicy(bindings=tuple())
        self.subjects = (
            pydantic.TypeAdapter(AuthenticatedSubject | AnonymousSubject)
        )
        inspect.markcoroutinefunction(self)
        assert inspect.iscoroutinefunction(self)

    async def authenticate(self, token: T) -> AuthenticatedSubject | AnonymousSubject:
        claims = token.model_dump(
            include=self.subject_claims
        )
        return self.subjects.validate_python(claims)

    async def authorize(
        self,
        request: fastapi.Request,
        subject: AuthenticatedSubject | AnonymousSubject
    ) -> AuthorizationContext:
        assert request.client
        ctx = AuthorizationContext.model_validate({
            'subject': subject,
            'remote_host': request.client.host
        })
        subject.roles.update(self.policy.granted(ctx))
        return ctx

    async def __call__(self, *args: Any, **kwargs: Any):
        return await self.authorize(*args, **kwargs)


import collections
import functools
import inspect
from inspect import Parameter
from typing import cast
from typing import Any
from typing import Callable
from typing import Iterable

import fastapi

from .types import ResourceServerException
from .types import InvalidRequestScope


class RequestHandler:
    requires: set[str]

    @functools.cached_property
    def __signature__(self):
        sig = inspect.signature(self.func)
        params = collections.OrderedDict(sig.parameters)
        params['_scope'] = Parameter(
            kind=Parameter.KEYWORD_ONLY,
            name='_scope',
            annotation=set[str],
            default=fastapi.Depends(self.get_scope)
        )
        return sig.replace(
            parameters=[
                Parameter(
                    kind=Parameter.POSITIONAL_ONLY,
                    name='_httprequest',
                    annotation=fastapi.Request,
                ),
                *list(params.values())
            ]
        )

    def __init__(
        self,
        func: Callable[..., Any]
    ):
        self.func = func
        self.requires = set()
        inspect.markcoroutinefunction(self)
        assert inspect.iscoroutinefunction(self)

    def get_scope(self, request: fastapi.Request) -> set[str]:
        return getattr(request.state, 'scope', set())

    def with_scope(self, scope: str):
        self.requires.add(scope)
        return self

    def validate_scope(self, request: fastapi.Request, granted: set[str]):
        missing = self.requires - granted
        if missing:
            raise InvalidRequestScope(missing)

    async def __call__(self, *args: Any, **kwargs: Any):
        request = cast(fastapi.Request, kwargs.pop('_httprequest'))
        assert getattr(request.state, 'subject', None)
        try:
            self.validate_scope(
                request=request,
                granted=set(cast(Iterable[str], kwargs.pop('_scope') or []))
            )
            return await self.func(*args, **kwargs)
        except ResourceServerException as e:
            return e.http()


import os
from typing import cast

import pydantic
from libcanonical.types import HTTPResourceLocator

from aegisx.ext.oauth.types import ResponseType
from aegisx.ext.oauth.types import ValidationContext
from ._fields import StateField
from ._fields import AuthorizationRequestScopeField
from ._oidcclaimspec import OIDCClaimSpec


class AuthorizationRequestParameters(pydantic.BaseModel):
    response_type: ResponseType = pydantic.Field(
        default=...,
        title="Response type",
        description=(
            "The desired response type from the authorization "
            "endpoint."
        )
    )

    client_id: str = pydantic.Field(
        default=...,
        title="Client ID",
        description=(
            "The client identifier as described in Section 2.2 of "
            "RFC 6749."
        )
    )

    redirect_uri: HTTPResourceLocator | str | None = pydantic.Field(
        default=None,
        title="Redirect URI",
        description=(
            "The client redirection endpoint as described in Section 3.1.2 "
            "of RFC 6749."
        )
    )

    state: StateField
    scope: AuthorizationRequestScopeField

    # OpenID Connect
    response_mode: str | None = pydantic.Field(
        default=None,
        title="Response mode",
        description=(
            " Informs the authorization server of the mechanism to be used for returning "
            "parameters from the authorization endpoint. This use of this parameter is NOT "
            "RECOMMENDED when the response mode that would be requested is the default mode "
            "specified for `response_type`."
        )
    )

    nonce: str | None = pydantic.Field(
        default=None,
        title="Nonce",
        description=(
            " String value used to associate a Client session with an ID Token, "
            "and to mitigate replay attacks. The value is passed through unmodified "
            "from the Authentication Request to the ID Token. If present in the ID "
            "Token, Clients MUST verify that the nonce Claim Value is equal to the "
            "value of the nonce parameter sent in the Authentication Request. "
            "If present in the Authentication Request, Authorization Servers MUST "
            "include a nonce Claim in the ID Token with the Claim Value being the "
            "nonce value sent in the Authentication Request. The nonce value is a "
            "case sensitive string"
        )
    )

    max_age: int | None = pydantic.Field(
        default=None,
        title="Maximum authentication age",
        description=(
            "Specifies the allowable elapsed time in seconds since the last time the "
            "end-user was actively authenticated by the authorization server. If the "
            "When max_age is used, the ID Token returned includes an `auth_time` claim. "
            "Note that `max_age=0` is equivalent to `prompt=login`."
        )
    )

    claims: OIDCClaimSpec | None = pydantic.Field(
        default=None,
        title="Claims",
        description=(
            "This parameter is used to request that specific claims be returned. The "
            "value is a JSON object listing the requested claims. The `claims` parameter "
            "value is represented in an OAuth 2.0 request as UTF-8 encoded JSON (which "
            "ends up being form-urlencoded when passed as an OAuth parameter). When used "
            "in a Request Object value, the JSON is used as the value of the `claims` member."
        )
    )

    @pydantic.model_validator(mode='after')
    def postprocess(self, info: pydantic.ValidationInfo):
        if not info.context:
            return self
        ctx = cast(ValidationContext, info.context)
        if self.is_openid() and ctx['op'] == 'create':
            self.nonce = bytes.hex(os.urandom(32))
        return self

    def is_openid(self):
        return self.scope and 'openid' in self.scope


import urllib.parse
from typing import Any

import pydantic

from ._codeauthorizationresponse import CodeAuthorizationResponse
from ._codeidtokentokenauthorizationresponse import CodeIDTokenTokenAuthorizationResponse
from ._codeidtokenauthorizationresponse import CodeIDTokenAuthorizationResponse
from ._codetokenauthorizationresponse import CodeTokenAuthorizationResponse
from ._idtokenauthorizationresponse import IDTokenAuthorizationResponse
from ._idtokentokenauthorizationresponse import IDTokenTokenAuthorizationResponse
from ._jarmauthorizationresponse import JARMAuthorizationServerResponse
from ._tokenauthorizationresponse import TokenAuthorizationResponse
from ._errorresponse import ErrorResponse


class AuthorizationResponse(
    pydantic.RootModel[
        JARMAuthorizationServerResponse |
        CodeIDTokenTokenAuthorizationResponse |
        IDTokenTokenAuthorizationResponse |
        IDTokenAuthorizationResponse |
        CodeTokenAuthorizationResponse |
        CodeIDTokenAuthorizationResponse |
        CodeAuthorizationResponse |
        TokenAuthorizationResponse |
        ErrorResponse
    ]
):

    @pydantic.model_validator(mode='before')
    @classmethod
    def preprocess_response(
        cls,
        response: Any | urllib.parse.ParseResult
    ) -> Any:
        if isinstance(response, urllib.parse.ParseResult):
            if response.query and response.fragment:
                raise ValueError(
                    "The response from the authorization endpoint must either "
                    "have a query or a fragment component, but not both."
                )
            if not response.query and not response.fragment:
                raise ValueError(
                    "The authorization server did not send a response in the "
                    "query or the fragment component."
                )
            match bool(response.query):
                case True:
                    try:
                        response = dict(urllib.parse.parse_qsl(response.query))
                    except ValueError:
                        raise ValueError(
                            "the query component of the authorization response "
                            "is malformed."
                        )
                case False:
                    raise NotImplementedError("Fragment responses are not implemented.")
        return response

    @property
    def code(self) -> str | None:
        return getattr(self.root, 'code', None)

    @property
    def iss(self) -> str | None:
        return getattr(self.root, 'iss', None)

    @property
    def redirect_uri(self) -> str | None:
        return getattr(self.root, 'redirect_uri', None)

    @property
    def state(self) -> str | None:
        return getattr(self.root, 'state', None)

    def fatal(self):
        if not isinstance(self.root, ErrorResponse):
            raise UserWarning(
                "Can not call AuthorizationResponse.fatal() on non-error "
                "responses."
            )
        self.root.fatal()

    def is_error(self):
        return self.root.is_error()


from ._resourceserverexception import ResourceServerException


class InvalidRequestToken(ResourceServerException):
    error = 'invalid_token'
    status_code = 403

    def __init__(self, message: str):
        super().__init__(message)


from ._baseresourceserver import BaseResourceServerAuth
from ._clientsecretcredential import ClientSecretCredential
from ._interactive import InteractiveAuth
from ._refreshtoken import RefreshTokenAuth


__all__: list[str] = [
    'BaseResourceServerAuth',
    'ClientSecretCredential',
    'InteractiveAuth',
    'RefreshTokenAuth',
]


from ._baseresourceserver import BaseResourceServerAuth


class RefreshTokenAuth(BaseResourceServerAuth):
    """A :class:`~BaseResourceServerAuth` implementation for use with a
    refresh token. It assumes that the grant specified by the `name`
    parameter is already present in the repository.
    """
    __module__: str = 'aegisx.ext.oauth.client.auth'


from typing import Iterable

import fastapi
from aegisx.ext.iam import AuthorizationContext

from aegisx.ext.oauth.resource.types import InvalidRequestScope
from aegisx.ext.oauth.resource.security import RequestContext
from ._currentscope import CurrentScope
from ._currentsubject import CurrentSubject


def Scope(scope: Iterable[str], *, context: RequestContext):
    required = set(scope)

    def require_scope(
        scope: CurrentScope,
        subject: CurrentSubject,
        context: AuthorizationContext = fastapi.Depends(context)
    ):
        missing = required - scope
        if missing:
            match subject.is_authenticated():
                case True:
                    raise InvalidRequestScope(missing).http()
                case False:
                    raise fastapi.HTTPException(
                        status_code=401,
                        headers={
                            'WWW-Authenticate': 'Bearer'
                        }
                    )

    return fastapi.Depends(require_scope)


from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

from aegisx.ext.jose import JSONWebToken
from aegisx.ext.jose import TokenValidator
from aegisx.ext.jose.models import JWSHeader

if TYPE_CHECKING:
    from ._accesstokenbearer import AccessTokenBearer


T = TypeVar('T', bound=Any, default=Any)


class AccessTokenValidator(TokenValidator[T]):

    def __init__(
        self,
        security: 'AccessTokenBearer[T]', 
        types: Any,
    ):
        self.security = security
        super().__init__(types=types)

    def is_accepted_issuer(self, payload: JSONWebToken) -> bool:
        result = self.security.is_accepted_issuer(payload) # type: ignore
        if result == NotImplemented:
            result = super().is_accepted_issuer(payload)
        assert isinstance(result, bool)
        return result

    async def get_verification_jwks(
        self,
        header: JWSHeader,
        payload: T
    ):
        return await self.security.get_verification_jwks(
            header=header,
            payload=payload,
            default=self.jwks
        )


from typing import Any

from ._httpsresourcelocator import HTTPSResourceLocator


class IssuerIdentifier(HTTPSResourceLocator):
    protocols = {'https'}

    @classmethod
    def validate(cls, v: str, _: Any = None):
        # RFC 9702 Section 2: The iss parameter value is the issuer
        # identifier of the authorization server that created the
        # authorization response, as defined in [RFC8414]. Its value
        # MUST be a URL that uses the "https" scheme without any query
        # or fragment components.
        self = super().validate(v, _)
        if self.parts.query:
            raise ValueError("The issuer identifier should not contain a query component.")
        if self.parts.fragment:
            raise ValueError("The issuer identifier should not contain a fragment component.")
        return self


import functools
from typing import overload
from typing import Any
from typing import Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aegisx.ext.oauth.models import ClientConfiguration
    from aegisx.ext.oauth.models import Grant


class IClientRepository:

    async def get(self, name: str) -> Union['ClientConfiguration', None]:
        ...

    async def grant(
        self,
        name: str
    ) -> Union['Grant', None]:
        ...

    @overload
    async def persist(self, obj: 'ClientConfiguration', *, name: str) -> None:
        ...

    @overload
    async def persist(self, obj: 'Grant', *, name: str, config: 'ClientConfiguration') -> None:
        ...

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: Union['ClientConfiguration', 'Grant'],
        *,
        name: str | None = None,
        **kwargs: Any
    ) -> None:
        ...


import inspect
import logging
import re
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generic
from typing import Literal
from typing import TypeVar

import fastapi
import fastapi.security
import pydantic
from aegisx.ext.iam import AnonymousSubject
from aegisx.ext.iam import AuthenticatedSubject
from aegisx.ext.jose import JSONWebKeySet
from aegisx.ext.jose import JSONWebToken
from aegisx.ext.jose import JWKSResolver
from aegisx.ext.jose import TokenValidator
from aegisx.ext.jose.models import JWSHeader
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded
from aegisx.ext.jose.types import Malformed
from aegisx.ext.jose.types import NotVerifiable
from aegisx.ext.jose.types import ForbiddenAudience
from aegisx.ext.jose.types import InvalidSignature
from aegisx.ext.jose.types import InvalidToken
from aegisx.ext.jose.types import Undecryptable
from aegisx.ext.jose.types import UntrustedIssuer

from aegisx.ext.oauth.resource.types import ResourceServerException
from aegisx.ext.oauth.resource.types import InvalidRequestToken
from ._accesstokenvalidator import AccessTokenValidator


Subject = AnonymousSubject | AuthenticatedSubject

T = TypeVar('T', default=JSONWebToken, bound=JSONWebToken)


class AccessTokenBearer(fastapi.security.HTTPBearer, Generic[T]):
    adapter: pydantic.TypeAdapter[JWECompactEncoded | JWSCompactEncoded]
    audience: set[str]
    audience_mode: Literal['domain', 'path']
    issuers: set[str] | re.Pattern[str] | None
    logger: logging.Logger = logging.getLogger(__name__)
    max_age: int
    subject_factory: Callable[[T], Subject | None | Awaitable[Subject | None]]
    subjects: set[str]
    types: T
    validator_class: type[AccessTokenValidator[T]] = AccessTokenValidator

    def __init__(
        self,
        types: Any,
        issuers: set[str] | re.Pattern[str] | None = None,
        audience: set[str] | None = None,
        audience_mode: Literal['domain', 'path'] = 'domain',
        max_age: int = 0,
        subjects: set[str] | None = None,
        subject_factory: Callable[[T], Subject | None | Awaitable[Subject | None]] | None = None,
        jwks_resolver: JWKSResolver | None = None,
        logger: logging.Logger | None = None
    ):
        super().__init__(
            auto_error=False,
            description=("OIDC ID Token")
        )
        self.adapter = pydantic.TypeAdapter(JWECompactEncoded | JWSCompactEncoded)
        self.audience = audience or set()
        self.audience_mode = audience_mode
        self.issuers = issuers
        self.jwks_resolver = jwks_resolver or JWKSResolver(domains=[])
        self.logger = logger or self.logger
        self.max_age = max_age
        self.subjects = subjects or set()
        self.subject_factory = subject_factory or (lambda _: AnonymousSubject()) # type: ignore
        self.types = types

    def get_audience(self, request: fastapi.Request) -> set[str]:
        return {*self.audience, *self.request_audience(request)}

    def get_token_validator(self, request: fastapi.Request) -> TokenValidator[T]:
        return self.validator_class(types=self.types, security=self)\
            .max_age(self.max_age)\
            .with_audiences(self.get_audience(request))\
            .with_subjects(self.subjects)

    def is_accepted_issuer(self, token: T) -> bool:
        if not isinstance(self.issuers, re.Pattern) or token.iss is None:
            return NotImplemented
        return bool(self.issuers.match(token.iss))

    def is_service_account(self, token: T) -> bool:
        return False

    def request_audience(self, request: fastapi.Request) -> set[str]:
        audience: set[str] = set()
        match self.audience_mode:
            case 'domain':
                audience.add(f'{request.url.scheme}://{request.url.netloc}')
            case 'path':
                audience.add(f'{request.url.scheme}://{request.url.netloc}{request.url.path}')
        return audience

    async def authenticate(
        self,
        request: fastapi.Request,
        bearer: fastapi.security.HTTPAuthorizationCredentials | None
    ):
        request.state.subject = AnonymousSubject()
        request.state.scope = set()
        if bearer is not None:
            try:
                token = await self.validate(request, bearer.credentials)
                request.state.scope = set(token.scope)
                request.state.subject = await self.get_subject(token)
                await self.validate_request(request, token, request.state.subject)
            except NotVerifiable:
                # We don't have any keys that can verify the signature and
                # there were no instruments to obtain them.
                raise InvalidRequestToken(
                    message=(
                        "The JSON Web Token (JWS) provided in the Authorization "
                        "header was signed using a key that is not trusted "
                        "by the server, and it had no means to obtain trusted "
                        "public keys."
                    )
                )
            except ForbiddenAudience:
                raise InvalidRequestToken(
                    message=(
                        'The intended audience specified by the "aud" claim '
                        'is not accepted by the server.'
                    )
                )
            except InvalidSignature:
                raise InvalidRequestToken(
                    message=(
                        'The credential was signed by an untrusted '
                        'key.'
                    )
                )
            except Undecryptable:
                raise InvalidRequestToken(
                    message=(
                        'The encrypted credential in the Authorization header '
                        'could not be decrypted with any known key. Consult the '
                        'service documentation on how to properly encrypt an '
                        'access token.'
                    )
                )
            except UntrustedIssuer as e:
                raise InvalidRequestToken(message=e.message)
            except (pydantic.ValidationError, InvalidToken, Malformed):
                raise InvalidRequestToken(
                    message=(
                        'The credential provided in the Authorization header '
                        'is malformed.'
                    )
                )
            except Exception as e:
                self.logger.exception(
                    'Caught fatal %s while validating credential: %s',
                    type(e).__name__,
                    repr(e)
                )
                raise InvalidRequestToken(
                    message=(
                        'The credential provided in the Authorization header '
                        'is not accepted.'
                    )
                )
        return request.state.subject

    async def get_subject(
        self,
        token: T
    ) -> AuthenticatedSubject | AnonymousSubject:
        subject = self.subject_factory(token)
        if inspect.isawaitable(subject):
            subject = await subject
        return subject or AnonymousSubject()

    async def get_verification_jwks(
        self,
        header: JWSHeader,
        payload: T,
        default: JSONWebKeySet
    ) -> JSONWebKeySet:
        jwks = default
        jku = payload.get_jwks_uri()
        if jku is not None:
            keys: set[str] = set()
            if header.kid:
                keys.add(header.kid)
            jwks = await self.jwks_resolver.resolve(jku, keys=keys)
        return jwks

    async def validate(
        self,
        request: fastapi.Request,
        token: JWECompactEncoded | JWSCompactEncoded | str
    ) -> T:
        if not isinstance(token, (JWECompactEncoded, JWSCompactEncoded)):
            token = self.adapter.validate_python(token)
        validator = self.get_token_validator(request)
        return await validator.validate(token)

    async def validate_request(
        self,
        request: fastapi.Request,
        token: T,
        subject: Subject
    ) -> None:
        pass

    async def __call__(self, request: fastapi.Request): # type: ignore
        request.state.subject = AnonymousSubject()
        bearer = await super().__call__(request)
        try:
            return await self.authenticate(request, bearer)
        except ResourceServerException as e:
            raise e.http()


import hashlib
import os
import stat

import yaml
from aegisx.core.const import AEGIS_USER_DIR

from aegisx.ext.oauth.models import ClientConfiguration
from aegisx.ext.oauth.models import Grant
from ._base import ClientRepository


AEGISX_CLIENT_DIR = AEGIS_USER_DIR.joinpath('clients.d')

AEGISX_GRANTS_DIR = AEGIS_USER_DIR.joinpath('grants.d')


class LocalClientRepository(ClientRepository):
    clients_dir = AEGISX_CLIENT_DIR
    grants_dir = AEGISX_GRANTS_DIR

    @staticmethod
    def client_name(name: str):
        return hashlib.sha256(name.encode()).hexdigest()

    @staticmethod
    def grant_name(name: str):
        return hashlib.sha256(name.encode()).hexdigest()

    def __init__(self):
        if not self.clients_dir.exists():
            os.makedirs(self.clients_dir)
        if not self.grants_dir.exists():
            os.makedirs(self.grants_dir)

    async def get(self, name: str) -> ClientConfiguration | None:
        filename = self.clients_dir.joinpath(self.client_name(name))
        if not filename.exists():
            return None
        with open(filename) as f:
            data = yaml.safe_load(f.read())
        return ClientConfiguration.model_validate(data)

    async def grant(self, name: str) -> Grant | None:
        filename = self.grants_dir.joinpath(f'{self.grant_name(name)}')
        if not filename.exists():
            return None
        with open(filename, 'r') as f:
            return Grant.model_validate_json(f.read())

    async def persist_client_config(
        self,
        obj: ClientConfiguration,
        *,
        name: str
    ) -> None:
        filename = self.clients_dir.joinpath(self.client_name(name))
        with open(filename, 'w') as f:
            f.write(obj.model_dump_json(exclude_none=True))
        os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)

    async def persist_grant(
        self,
        obj: Grant,
        *,
        name: str,
        config: ClientConfiguration
    ) -> None:
        filename = self.grants_dir.joinpath(self.grant_name(name))
        with open(filename, 'w') as f:
            f.write(obj.model_dump_json(exclude_none=True))
        os.chmod(filename, stat.S_IRUSR | stat.S_IWUSR)




#: The set of standard claims included in an ID Token per
#: OpenID Connect Core 1.0.
OICD_STANDARD_CLAIMS: set[str] = {
    'iss', 'sub', 'aud', 'exp', 'iat'
}

OIDC_AUTHORIZATION_CODE_FLOW_CLAIMS: set[str] = {'at_hash'}

OIDC_IMPLICIT_FLOW_CLAIMS: set[str] = {'at_hash', 'nonce'}

OIDC_HYBRID_FLOW_CLAIMS: set[str] = {'at_hash', 'c_hash', 'nonce'}


from typing import Annotated

import fastapi
from aegisx.ext.iam import AuthenticatedSubject
from aegisx.ext.iam import AnonymousSubject


async def get(request: fastapi.Request):
    subject = getattr(request.state, 'subject', None)
    if subject is None:
        subject = AnonymousSubject()
    return subject


CurrentSubject = Annotated[
    AuthenticatedSubject | AnonymousSubject,
    fastapi.Depends(get)
]


from typing import ClassVar
from typing import Generic
from typing import NoReturn
from typing import TypeVar

import pydantic
from libcanonical.types import HTTPResourceLocator

from aegisx.ext.oauth.types import Error
from ._authorizationserverresponse import AuthorizationServerResponse
from ._fields import IssuerIdentifierField
from ._fields import StateField


E = TypeVar('E', default=str)


class ErrorResponse(AuthorizationServerResponse, Generic[E]):
    model_config = {'extra': 'ignore'}
    code: ClassVar[None] = None

    error: E = pydantic.Field(
        default=...,
        title="Error",
        description=(
            "The error code returned from the server."
        )
    )

    error_description: str | None = pydantic.Field(
        default=None,
        title="Description",
        description=(
            "Human-readable ASCII text providing additional information, used "
            "to assist the client developer in understanding the error that "
            "occurred. Values for the `error_description` parameter MUST NOT "
            "include characters outside the set `%x20-21 / %x23-5B / %x5D-7E`."
        )
    )

    error_uri: HTTPResourceLocator | str | None = pydantic.Field(
        default=None,
        title="URI",
        description=(
            "A URI identifying a human-readable web page with information about the error, "
            "used to provide the client developer with additional information about the "
            "error. Values for the `error_uri` parameter MUST conform to the URI-reference "
            "syntax and thus MUST NOT include characters outside the set `%x21 / %x23-5B / "
            "%x5D-7E`."
        )
    )

    state: StateField
    iss: IssuerIdentifierField

    def fatal(self) -> NoReturn:
        Error(self.error_description).fatal()

    def is_error(self):
        return True

    def throw(self) -> NoReturn:
        raise Error(self.error_description)


import functools
from typing import overload
from typing import Any

from aegisx.ext.oauth.protocols import IClientRepository
from aegisx.ext.oauth.models import ClientConfiguration
from aegisx.ext.oauth.models import Grant
from aegisx.ext.oauth.models import TokenResponse


class ClientRepository(IClientRepository):

    async def grant(
        self,
        name: str
    ) -> Grant | None:
        """Lookup a grant from the persistent data storage."""
        raise NotImplementedError

    async def get(self, name: str) -> ClientConfiguration | None:
        raise NotImplementedError

    @overload
    async def persist(self, obj: ClientConfiguration, *, name: str) -> None:
        ...

    @overload
    async def persist(self, obj: Grant, *, name: str, config: ClientConfiguration) -> None:
        ...

    @functools.singledispatchmethod
    async def persist(
        self,
        obj: ClientConfiguration | TokenResponse,
        *,
        name: str | None = None,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError(f'Can not persist {type(obj).__name__}')

    async def persist_client_config(self, obj: ClientConfiguration, *, name: str) -> None:
        raise NotImplementedError

    async def persist_grant(self, obj: Grant, *, name: str, config: ClientConfiguration) -> None:
        raise NotImplementedError

    @persist.register # type: ignore
    async def _(self, obj: ClientConfiguration, *, name: str):
        return await self.persist_client_config(obj, name=name)

    @persist.register # type: ignore
    async def _(self, obj: Grant, name: str, config: ClientConfiguration):
        return await self.persist_grant(obj, name=name, config=config)


import functools
import os
import secrets
import urllib.parse
from typing import Any
from typing import Callable
from typing import TypeVar

import httpx

from aegisx.ext.oauth.models import AuthorizationResponse
from aegisx.ext.oauth.models import AuthorizationRequestParameters
from aegisx.ext.oauth.models import ClientConfiguration
from aegisx.ext.oauth.models import OIDCToken
from aegisx.ext.oauth.models import ServerMetadata
from aegisx.ext.oauth.models import TokenRequest
from aegisx.ext.oauth.models import TokenResponse
from aegisx.ext.oauth.types import IssuerIdentifier
from aegisx.ext.oauth.types import ClientAuthenticationMethod
from aegisx.ext.oauth.types import ClientConfigurationError
from aegisx.ext.oauth.types import HTTPSResourceLocator
from aegisx.ext.oauth.types import NeedsDiscovery
from aegisx.ext.oauth.types import ResponseType
from .auth import ClientSecretCredential
from ._oidctokenvalidator import OIDCTokenValidator

R = TypeVar('R')
U = TypeVar("U", bound="Client")


class Client(httpx.AsyncClient):

    @property
    def client_id(self):
        assert self.credential
        return self.credential.client_id

    @staticmethod
    def needs_credential(func: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(func)
        def f(self: 'Client', *args: Any, **kwargs: Any) -> R:
            if not self.credential:
                raise ValueError(f"{type(self).__name__}.credential can not be None.")
            return func(self, *args, **kwargs)
        return f

    @staticmethod
    def needs_discovery(func: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(func)
        def f(self: 'Client', *args: Any, **kwargs: Any) -> R:
            metadata = kwargs.setdefault('metadata', self.metadata)
            if not metadata.is_discovered():
                raise NeedsDiscovery
            return func(self, *args, **kwargs)
        return f

    @staticmethod
    def needs_param(name: str):
        def decorator_factory(func: Callable[..., R]):
            @functools.wraps(func)
            def f(self: 'Client', *args: Any, **kwargs: Any):
                metadata = kwargs.get('metadata') or self.metadata
                if not hasattr(metadata, name):
                    raise AttributeError(f"ServerMetadata has no attribute {name}")
                if getattr(metadata, name) is None:
                    raise ClientConfigurationError(
                        f"The \"{name}\" parameter was not discovered or explicitely "
                        "configured."
                    )
                return func(self, *args, **kwargs)
            return f
        return decorator_factory

    @classmethod
    def fromconfig(cls, config: ClientConfiguration):
        return cls(
            issuer=config.issuer,
            credential=ClientSecretCredential(
                client_id=config.client_id,
                client_secret=config.client_secret,
                method='client_secret_post'
            ),
            metadata=config.metadata
        )

    @classmethod
    def fromservermetadata(cls, credential: ClientSecretCredential, metadata: ServerMetadata):
        return cls(
            issuer=IssuerIdentifier(metadata.issuer),
            credential=credential,
            metadata=metadata
        )

    def __init__(
        self,
        *,
        issuer: IssuerIdentifier | str,
        credential: ClientSecretCredential | None = None,
        authorization_endpoint: str | None = None,
        token_endpoint: str | None = None,
        metadata: ServerMetadata | None = None,
        oidc_validator: Callable[['Client', 'ServerMetadata', AuthorizationRequestParameters, 'TokenResponse'], OIDCTokenValidator] = OIDCTokenValidator,
        discoverable: bool = False
    ):
        self.credential = credential
        self.discoverable = discoverable
        self.oidc_validator = oidc_validator

        # TODO: Add commonly used parameters for now. For any extended features
        # the caller must provide the `metadata` parameter.
        manual_discovery = (not metadata or not metadata.is_discovered()) and not discoverable
        self.metadata = metadata or ServerMetadata.model_validate({
            'issuer': issuer,
            'authorization_endpoint': authorization_endpoint,
            'token_endpoint': token_endpoint,
        })
        if manual_discovery:
            self.metadata.discovered = True
            assert self.metadata.is_discovered()
        super().__init__(
            timeout=60.0,
            verify=not os.getenv('AEGISX_SSL_VERIFY') == '0'
        )

    @needs_discovery
    @needs_param('authorization_endpoint')
    def authorize_url(
        self,
        response_type: ResponseType | str,
        redirect_uri: str | None = None,
        state: str | None = None,
        scope: set[str] | None = None,
        *,
        metadata: ServerMetadata | None = None,
        **kwargs: Any,
    ) -> HTTPSResourceLocator:
        assert self.credential
        metadata = metadata or self.metadata
        if scope is not None:
            kwargs['scope'] = scope
        params = AuthorizationRequestParameters.model_validate(
            {
                'client_id': self.credential.client_id,
                'response_type': response_type,
                'redirect_uri': redirect_uri,
                'state': state,
                **kwargs
            },
            context={'op': 'create'}
        )
        q = params.model_dump(
            mode='json',
            exclude_none=True,
            exclude_defaults=True,
            exclude_unset=True
        )
        assert metadata.authorization_endpoint
        return params, metadata.authorization_endpoint.with_query(**q) # type: ignore

    async def connect(self):
        if not self.metadata.is_discovered() and self.discoverable:
            await self.metadata.discover(client=self)
        return self

    @needs_discovery
    @needs_param('token_endpoint')
    async def obtain(
        self,
        request: AuthorizationRequestParameters,
        response: AuthorizationResponse,
        metadata: ServerMetadata
    ):
        """Obtain a new access token using an authorization code.
        
        Args:
            request (AuthorizationRequestParameters): the parameters of the
                authorization request that produced the response.
            response (AuthorizationResponse): the response from the authorization
                endpoint.
        """
        assert self.credential
        assert metadata.token_endpoint
        if not response.code:
            raise ValueError(
                f"{type(response.root).__name__} does not supply an "
                "authorization code."
            )

        # The "state" parameter in the Authorization Response MUST
        # match the "state" parameter that was included in the
        # Authorization Request.
        if request.state:
            if not response.state:
                raise ValueError(
                    "The authorization server did not return the "
                    "\"state\" parameter."
                )
            if not secrets.compare_digest(request.state, response.state):
                raise ValueError(
                    "The \"state\" parameter returned by the authorization "
                    "server did not match with the local value."
                )

        # Clients that support this specification MUST extract the value of
        # the iss parameter from authorization responses they receive if the
        # parameter is present. Clients MUST compare the extracted and URL-
        # decoded value to the issuer identifier of the authorization server
        # where the authorization request was sent to. This comparison MUST
        # use simple string comparison as defined in Section 6.2.1. of [RFC3986].
        # If the value does not match the expected issuer identifier, clients
        # MUST reject the authorization response and MUST NOT proceed with
        # the authorization grant.
        if response.iss:
            if not secrets.compare_digest(response.iss, metadata.issuer):
                raise ValueError(
                    "The \"iss\" parameter sent by the authorization "
                    "server does not match the known metadata: "
                    f"{metadata.issuer}"
                )

        # The request and response are valid, proceed to
        # obtain the token. If the server issued an ID
        # token, validate the token according to the
        # OpenID Connect Core specification.
        grant = TokenRequest.model_validate({
            'grant_type': 'authorization_code',
            'code': response.code,
            'redirect_uri': request.redirect_uri
        })
        self.credential.add_to_grant(grant)
        token = await self._grant(metadata.token_endpoint, grant)
        if token.id_token:
            validator = self.oidc_validator(self, metadata, request, token)
            await token.validate_id_token(validator)
        return token

    async def on_redirected(
        self,
        result: urllib.parse.ParseResult,
        state: str | None = None
    ):
        response = AuthorizationResponse.model_validate(obj=result)
        return response

    @needs_discovery
    @needs_param('token_endpoint')
    async def refresh(
        self,
        refresh_token: str,
        metadata: ServerMetadata
    ) -> TokenResponse:
        assert self.credential
        assert metadata.token_endpoint
        request = TokenRequest.model_validate({
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        })
        self.credential.add_to_grant(request)
        return await self._grant(metadata.token_endpoint, request)

    @needs_discovery
    @needs_param('userinfo_endpoint')
    async def userinfo(
        self,
        token: str,
        metadata: ServerMetadata | None,
        endpoint: str | None = None
    ) -> OIDCToken:
        assert metadata is not None
        assert metadata.userinfo_endpoint is not None
        response = await self.get(
            url=endpoint or metadata.userinfo_endpoint,
            headers={'Authorization': f'Bearer {token}'}
        )
        return OIDCToken.model_validate(response.json())

    async def __aenter__(self: U) -> U:
        await super().__aenter__()
        await self.connect()
        return self

    async def _grant(self, endpoint: str, grant: TokenRequest):
        self._select_client_secret_jwt_alg(self.metadata)
        response = await self.post(
            url=endpoint,
            data=grant.model_dump(
                mode='json',
                exclude_unset=True,
                exclude_defaults=True,
                exclude_none=True
            )
        )
        try:
            return TokenResponse.model_validate(
                response.json(),
                context=response
            )
        except Exception:
            raise NotImplementedError(f"invalid response from authorization server:\n\n{response.text}.")

    def _select_client_secret_jwt_alg(self, metadata: ServerMetadata):
        assert self.credential
        if self.credential.method != ClientAuthenticationMethod.client_secret_jwt:
            return
        if self.credential.alg:
            return
        candidates = [
            x for x in
            metadata.token_endpoint_auth_signing_alg_values_supported or []
            if str.startswith(x, 'HS')
        ]
        if not candidates:
            raise ValueError(
                "Unable to select a suitable algorithm to use with "
                "client authentication method client_secret_jwt from "
                f"algorithms: {metadata.token_endpoint_auth_signing_alg_values_supported}"
            )
        self.credential.alg = candidates[0]


import secrets

import fastapi


class ResourceServerException(Exception):
    error: str
    status_code: int

    @staticmethod
    def escape(value: str):
        return value.replace('\\', '\\\\').replace('"', '\\"')

    def __init__(self, message: str, /):
        self.message = message

    def get_www_authenticate_header(self) -> str:
        return f'Bearer, error="{self.error}", error_description: "{self.escape(self.message)}"'

    def http(self):
        header = self.get_www_authenticate_header()
        header = f'{header}, nonce="{secrets.token_urlsafe(32)}"'
        raise fastapi.HTTPException(
            status_code=self.status_code,
            detail=self.message,
            headers={'WWW-Authenticate': header}
        )


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from aegisx.ext.oauth.client import OIDCTokenValidator

import pydantic
from aegisx.ext.jose import JSONWebToken
from aegisx.ext.jose.types import JWECompactEncoded
from aegisx.ext.jose.types import JWSCompactEncoded

from aegisx.ext.oauth.types import AccessTokenType
from ._accesstokentokenresponse import AccessTokenTokenResponse
from ._errorresponse import ErrorResponse
from ._nonprotocolerrorresponse import NonProtocolErrorResponse
from ._proofofpossessiontokenresponse import ProofOfPossessionTokenResponse


class TokenResponse(
    pydantic.RootModel[
        ProofOfPossessionTokenResponse |
        AccessTokenTokenResponse |
        ErrorResponse |
        NonProtocolErrorResponse
    ]
):

    @property
    def access_token(self) -> str | None:
        return getattr(self.root, 'access_token', None)

    @property
    def expires_in(self) -> int | None:
        return getattr(self.root, 'expires_in', None)

    @property
    def id_token(self) -> JWECompactEncoded | JWSCompactEncoded | None:
        return getattr(self.root, 'id_token', None)

    @property
    def jwt(self) -> JSONWebToken | None:
        return getattr(self.root, 'jwt', None)

    @property
    def refresh_token(self) -> str | None:
        return getattr(self.root, 'refresh_token')

    @property
    def token_type(self) -> AccessTokenType:
        return getattr(self.root, 'token_type')

    def fatal(self):
        assert isinstance(self.root, (ErrorResponse, NonProtocolErrorResponse))
        self.root.fatal()

    def is_encrypted(self):
        return self.root.is_encrypted()

    def is_error(self):
        return isinstance(self.root, (ErrorResponse, NonProtocolErrorResponse))

    def is_signed(self):
        return self.root.is_signed()

    async def validate_id_token(self, validator: 'OIDCTokenValidator'):
        assert isinstance(self.root, AccessTokenTokenResponse)
        return await self.root.validate_id_token(validator)


from ._client import Client
from ._oidctokenvalidator import OIDCTokenValidator
from .auth import *
from .repo import *


__all__: list[str] = [
    'Client',
    'ClientRepository',
    'ClientSecretCredential',
    'InteractiveAuth',
    'OIDCTokenValidator',
]


from typing import TYPE_CHECKING

import httpx
from aegisx.ext.jose import JSONWebKey

from aegisx.ext.oauth.types import ClientAuthenticationMethod
if TYPE_CHECKING:
    from aegisx.ext.oauth.models import TokenRequest


class ClientSecretCredential(httpx.Auth):
    alg: str | None = None

    @property
    def jwk(self) -> JSONWebKey | None:
        if not self.client_secret:
            return None
        return JSONWebKey.model_validate({
            'kty': 'oct',
            'alg': self.alg or 'HS',
            'k': self.client_secret
        })

    def __init__(
        self,
        client_id: str,
        client_secret: str | None,
        *,
        method: ClientAuthenticationMethod | str = ClientAuthenticationMethod.none,
        alg: str | None = None
    ):
        if isinstance(method, str): # type: ignore
            method = ClientAuthenticationMethod(method)
        self.alg = alg
        self.client_id = client_id
        self.client_secret = client_secret
        self.method = method

    def add_to_grant(self, grant: 'TokenRequest'):
        if self.method == ClientAuthenticationMethod.client_secret_post:
            grant.root.client_id = self.client_id
            grant.root.client_secret = self.client_secret


from ._authorizationserverresponse import AuthorizationServerResponse
from ._fields import IssuerIdentifierField
from ._fields import StateField


class AuthorizationResponseBase(AuthorizationServerResponse):
    model_config = {'extra': 'forbid'}
    state: StateField
    iss: IssuerIdentifierField


import urllib.parse
from typing import Union

import pydantic

from ._authorizationcodegrant import AuthorizationCodeGrant
from ._refreshtokenrequest import RefreshTokenRequest


class TokenRequest(
    pydantic.RootModel[
        Union[
            AuthorizationCodeGrant,
            RefreshTokenRequest
        ]
    ]
):
    
    @pydantic.model_serializer(mode='wrap')
    def serialize(
        self,
        nxt: pydantic.SerializerFunctionWrapHandler,
        info: pydantic.SerializationInfo
    ):
        if info.mode != 'form': # type: ignore
            return nxt(self)
        return urllib.parse.urlencode(
            self.model_dump(
                mode='json',
                exclude_defaults=True,
                exclude_unset=True,
                exclude_none=True
            )
        )


from typing import Any

import pydantic

from ._oidcclaimrequest import ClaimRequest


class RequestedClaims(pydantic.BaseModel):
    model_config = {'extra': 'allow'}
    __pydantic_extra__: dict[str, ClaimRequest[int | str]] = pydantic.Field( # type: ignore
        init=False
    )

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        for key, value in values.items():
            if value is not None:
                continue
            values[key] = ClaimRequest(essential=False)
        return values

    def requested(self) -> set[str]:
        """Return the set of claims that are requested."""
        return set(self.__pydantic_extra__.keys())


from ._resourceserverexception import ResourceServerException


class InvalidRequestScope(ResourceServerException):
    error = 'insufficient_scope'
    status_code = 403

    def __init__(self, scope: set[str], message: str | None = None):
        super().__init__(
            'The request requires higher privileges than granted '
            'by the access token.'
        )
        self.scope = scope

    def get_www_authenticate_header(self) -> str:
        header = super().get_www_authenticate_header()
        return f'{header}, scope="{" ".join(sorted(map(self.escape, self.scope)))}"'